/*
Author: Gao Mingcen & Cao Thanh Tung
Date: 28/02/2013

File Name: main.cpp

Example of the usage of ffHull (WITH VISUALIZATION)

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#pragma warning(disable: 4018 4101 4715 4996)

#include <stdio.h>
#include <stdlib.h>

#include "FFHull.h"
#include "PointsCreater.h"
#include "GLFunction.h"

using namespace std;

int main(int argc, char *argv[]) //for 3D convex hull
{
	FFHull::GSetDevice(0);

	int mode = 1; //0:cpu 1: gpu
	int seednum = 10000000;

	//Initialize a mesh, thisMesh.
	TriangleMesh3D * thisMesh = new TriangleMesh3D();
	thisMesh->initialize(seednum);

	//Randomly create a point set for thisMesh
	createPoint3D(thisMesh->points, seednum, 1.0);

	PerformanceTimer pt;
	pt.StartTimer();

	//Initialize an instance of FFHull with the mesh
	FFHull * thisFlipper = new FFHull(thisMesh);
	
	//Compute the convex hull of the point set stored in thisMesh
	//The first parameter indicates the mode of FFHull. Use 0 for CPU computation and 1 for GPU computation.
	//The second parameter indicates whether to show the detail of the computation
	//You can also use ComputeConvexHull() or ComputeConvexHull(_mode). By default FFHull runs in CPU without computation details.
	thisFlipper->ComputeConvexHull(mode, true);

	printf("Computing Time: %.3f ms\n\n", pt.GetTimeElapsed() * 1000);

	//Visualize the mesh 
	gl_visualize(thisMesh);

	//Clear mesh
	thisMesh->clear();
	free(thisMesh);

	return 0;
}