/*
Author: Gao Mingcen & Cao Thanh Tung
Date: 28/02/2013

File Name: FFHull.h

Class Definition of FFHull

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#pragma once
#include "common.h"
#include "GpuMemoryManager.h"
#include "GpuPrimitiveManager.h"
#include <queue>

typedef struct {
    int fromPoint, toPoint; 
    int triangle; 
} StackItem; 

class FFHull
{
private:
	
	//basic variables
	static const int PredParaNum = 14;	// the size of hostConst
	REAL hostConst[14];					// constants used in exact CUDA CCW test

	int mode; //-1: empty; 0: cpu; 1: gpu
	bool growing;
	bool detail;
	int iniPoint[4];
	int * triToPoint;	//store one special point associated to the triangle
	REAL * triToValue;	//store one special value associated with triToPoint
	int triArraySize;

	//basic functions
	void g_exactInit();
	void c_exactInit();
	void _setMode(int _mode);
	void _unsetMode();
	
	//variables and functions for the cpu version
	int * pointOrder; 
    char * pointLoc; 
	vector<Edge> cCheckEdge; 
	vector<StackItem> cStack; 
	void c_Initialization();
	void c_Deinitialization();
	void c_ComputeTetra();
	void c_GrowHull();
	void c_CorrectHull();
	void c_CompactMesh();
	bool c_assignLayer(const StackItem item);
	void c_subdivide(const StackItem item);
	void c_pFlipEdge(const Edge edge);
	void c_flip31(int triCodeLeft, int triCodeRight);
	void c_flip22(int triCodeLeft, int triCodeRight);
	void c_correctInitialGroup();
	int c_locateIniTri(int pid, REAL & pred);
	void c_expandArray(void ** data, int oldLength, int newLength); //length in byte
	int c_posInTriangle(Triangle * tri, int pid);
	void c_clearTriangle(Triangle & tri);
	REAL c_check3DOrientation(TriangleMesh3D * tMesh, int i0, int i1, int i2, int i3);
	REAL c_check3DOrientationSOS(TriangleMesh3D * tMesh, int i0, int i1, int i2, int i3);
    void c_reorderPoints( int *pointOrder, char *pointLoc, int *offset, int noVal, int beg, int end);

	//variables and functions for the gpu version
	int * g_pointIndex;
	int * g_triPrefix;
	int g_noAlive;
	GpuPrimitiveManager * primitiveManager;
	GpuMemoryManager * memoryManager;
	TriangleMesh3D *cmesh;
	TriangleMesh3D *gmesh;
	void g_Initialization();
	void g_Deinitialization();
	void g_ComputeTetra();
	void g_GrowHull();
	void g_CorrectHull();
	void g_CompactMesh();
	void g_assignLayer();
	void g_subdivide();
	void g_copyMesh(TriangleMesh3D * host_mesh);

public:
	FFHull(TriangleMesh3D * tMesh);
	~FFHull(void);
	void Initialization();
	void Deinitialization();
	TriangleMesh3D * ComputeConvexHull();
	TriangleMesh3D * ComputeConvexHull(int _mode);
	TriangleMesh3D * ComputeConvexHull(int _mode, bool _detail);
	void ComputeTetra();
	void GrowHull();
	void CorrectHull();
	void CompactMesh();	
	static void GSetDevice(int id);
};
