/*
Author: Gao Mingcen & Cao Thanh Tung
Date: 28/02/2013

File Name: FFHull_Kernel.h

The GPU kernels for FFHull, refered by FFHull_CPU.cu 

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#include "gpuPredicates.h"

__device__ REAL g_check3DOrientationSOS(Point3D point0, Point3D point1, Point3D point2, Point3D eye,
									 int i0, int i1, int i2, int i3)
{
	int depth = 0, pn = 1;
	REAL p3[4][3];
	int index[4];
	REAL pa2[2], pb2[2], pc2[2];

	p3[0][0] = point0.x;
	p3[1][0] = point1.x;
	p3[2][0] = point2.x;
	p3[3][0] = eye.x;
	p3[0][1] = point0.y;
	p3[1][1] = point1.y;
	p3[2][1] = point2.y;
	p3[3][1] = eye.y;
	p3[0][2] = point0.z;
	p3[1][2] = point1.z;
	p3[2][2] = point2.z;
	p3[3][2] = eye.z;

	REAL result = -orient3d(p3[0], p3[1], p3[2], p3[3]);

	if(result == 0)
	{
		index[0] = i0; index[1] = i1; index[2] = i2; index[3] = i3;

        for (int i = 0; i < 3; i++) {
            int min = i; 

            for (int j = i + 1; j < 4; j++) 
                if (index[j] < index[min]) 
                    min = j; 

            if (min != i) {
                int tempI = index[min]; 
                index[min] = index[i]; 
                index[i] = tempI; 

                for (int m = 0; m < 3; m++) {
                    REAL tempR = p3[min][m]; 
                    p3[min][m] = p3[i][m]; 
                    p3[i][m] = tempR; 
                }

                pn = -pn; 
            }
        }

        //loop mode
		for(depth = 0; depth < 14; depth ++)
		{
			switch(depth)
			{
			case 0:
				pa2[0] = p3[1][0];
				pa2[1] = p3[1][1];
				pb2[0] = p3[2][0];
				pb2[1] = p3[2][1];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][1];
				break;
			case 1:
				pa2[0] = p3[1][0];
				pa2[1] = p3[1][2];
				pb2[0] = p3[2][0];
				pb2[1] = p3[2][2];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][2];
				break;
			case 2:
				pa2[0] = p3[1][1];
				pa2[1] = p3[1][2];
				pb2[0] = p3[2][1];
				pb2[1] = p3[2][2];
				pc2[0] = p3[3][1];
				pc2[1] = p3[3][2];
				break;
			case 3:
				pa2[0] = p3[0][0];
				pa2[1] = p3[0][1];
				pb2[0] = p3[2][0];
				pb2[1] = p3[2][1];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][1];
				break;
			case 4:
				result = p3[2][0] - p3[3][0];
				break;
			case 5:
				result = p3[2][1] - p3[3][1];
				break;
			case 6:
				pa2[0] = p3[0][0];
				pa2[1] = p3[0][2];
				pb2[0] = p3[2][0];
				pb2[1] = p3[2][2];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][2];
				break;
			case 7:
				result = p3[2][2] - p3[3][2];
				break;
			case 8:
				pa2[0] = p3[0][1];
				pa2[1] = p3[0][2];
				pb2[0] = p3[2][1];
				pb2[1] = p3[2][2];
				pc2[0] = p3[3][1];
				pc2[1] = p3[3][2];
				break;
			case 9:
				pa2[0] = p3[0][0];
				pa2[1] = p3[0][1];
				pb2[0] = p3[1][0];
				pb2[1] = p3[1][1];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][1];
				break;
			case 10:
				result = p3[1][0] - p3[3][0];
				break;
			case 11:
				result = p3[1][1] - p3[3][1];
				break;
			case 12:
				result = p3[0][0] - p3[3][0];
				break;
			default:
				result = 1.0;
				break;
			}

			if(depth < 4 || depth ==6 || depth == 8 || depth == 9)
				result = orient2d(pa2, pb2, pc2);
			if(result != 0)
			{
				if(depth == 1 || depth == 3 || depth == 5 || depth == 8 || depth == 10)
					result = - result;
				break;
			}
		}
	}

	return result * pn;
}
__device__ REAL g_check3DOrientation(Point3D point0, Point3D point1, Point3D point2, Point3D eye,
									 int i0, int i1, int i2, int i3)
{
	//1: Without SOS use our own logic, it is currently unstable
	//return -orient3d((REAL *)&point0, (REAL *)&point1, (REAL *)&point2, (REAL *)&eye);

	//2: Herbert's SOS provides a stable version
	return g_check3DOrientationSOS(point0, point1, point2, eye, i0,i1,i2,i3);
}
__device__ REAL g_check3DOrientationExact(Point3D point0, Point3D point1, Point3D point2, Point3D eye)
{
	return -orient3dexact((REAL *)&point0, (REAL *)&point1, (REAL *)&point2, (REAL *)&eye);
}

__device__ REAL g_check3DOrientationFast(Point3D point0, Point3D point1, Point3D point2, Point3D eye, bool &exact)
{
	return -orient3dfast((REAL *)&point0, (REAL *)&point1, (REAL *)&point2, (REAL *)&eye, exact);
}

__device__ int cd_LocateIniTriFast(Point3D * points, Point3D kernel, 
								   int pid, int a, int b, int c, int d, REAL & pred, bool & exact)
{
	exact = true;
	Point3D iniPoint[4] = {points[a], points[b], points[c], points[d]};
	Point3D point = points[pid];
	pred = 0.0;
	int result = -1;
	REAL res30, res31, res32, res;
	int ip1, ip2, ip3;
	res31 = g_check3DOrientationFast(kernel, iniPoint[3], iniPoint[1], point, exact);
	res32 = g_check3DOrientationFast(kernel, iniPoint[3], iniPoint[2], point, exact);
	if((res31<0 && res32>0) || res31 == 0 || res32 == 0) //pick the smaller triangle
	{result = 1;ip1 = 1; ip2= 2;}
	else
	{			
		res30 = g_check3DOrientationFast(kernel, iniPoint[3], iniPoint[0], point, exact);
		if((res32<0 && res30>0) || res30 == 0)
		{result = 2;ip1 = 2; ip2= 0;}
		else if(res30<0 && res31>0)
		{result = 3;ip1 = 0; ip2= 1;}
	}
	res = g_check3DOrientationFast(kernel, iniPoint[ip1], iniPoint[ip2], point, exact);
	if(res >= 0) //below the plane, or on the boundary
	{result = 0; ip3 = Next(ip2);}
	else //above the plane
	{ip3 = ip2; ip2 = 3;}
	pred = g_check3DOrientationFast(iniPoint[ip1], iniPoint[ip2], iniPoint[ip3], point, exact);
	if(pred <=0) //means inside the tetrahedral
		result = -1;
	return result;
}
__global__ void cudaCorrectInitialGroupFast(Point3D * points, int * pointMap, Point3D kernel, int * triToPoint, 
											REAL * triToValue, int pointNum, int a, int b, int c, int d)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	REAL res;
	int curTriId = -1;
	REAL value;
	bool exact = true;

	while(id < pointNum)
	{
		if(id != a && id != b && id != c && id != d)
		{
			curTriId = cd_LocateIniTriFast(points, kernel, id, a, b, c, d, res, exact);
			if(!exact) //set the point to 2
			{
				curTriId = -2;
			}
			//select the maximum one
			else if(curTriId >= 0) //then res == 0, the point is on the triangle
			{			
				value = triToValue[curTriId];
				if(res > value)
				{
					triToValue[curTriId] = res;
					triToPoint[curTriId] = id;
				}
			}
		}
		pointMap[id] = curTriId;
		id += total;
	}
}
__device__ int cd_LocateIniTriExact(Point3D * points, Point3D kernel, int pid, int a, int b, int c, int d, REAL & pred)
{
	Point3D iniPoint[4] = {points[a], points[b], points[c], points[d]};
	Point3D point = points[pid];
	pred = 0.0;
	int result = -1;
	REAL res30, res31, res32, res;
	int ip1, ip2, ip3;
	res31 = g_check3DOrientationExact(kernel, iniPoint[3], iniPoint[1], point);
	res32 = g_check3DOrientationExact(kernel, iniPoint[3], iniPoint[2], point);
	if((res31<0 && res32>0) || res31 == 0 || res32 == 0) //pick the smaller triangle
	{result = 1;ip1 = 1; ip2= 2;}
	else
	{			
		res30 = g_check3DOrientationExact(kernel, iniPoint[3], iniPoint[0], point);
		if((res32<0 && res30>0) || res30 == 0)
		{result = 2;ip1 = 2; ip2= 0;}
		else if(res30<0 && res31>0)
		{result = 3;ip1 = 0; ip2= 1;}
	}
	res = g_check3DOrientationExact(kernel, iniPoint[ip1], iniPoint[ip2], point);
	if(res >= 0) //below the plane, or on the boundary
	{result = 0; ip3 = Next(ip2);}
	else //above the plane
	{ip3 = ip2; ip2 = 3;}
	pred = g_check3DOrientationExact(iniPoint[ip1], iniPoint[ip2], iniPoint[ip3], point);
	if(pred <=0) //means inside the tetrahedral
		result = -1;
	return result;
}
__global__ void cudaCorrectInitialGroupExact(Point3D * points, int * pointMap, Point3D kernel,
							int * triToPoint, REAL * triToValue, int pointNum, int a, int b, int c, int d)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	REAL res;
	int curTriId = -1;
	REAL value;

	while(id < pointNum)
	{
		if(pointMap[id] == -2) //means the first fast check give non-exact result
		{
			curTriId = cd_LocateIniTriExact(points, kernel, id, a, b, c, d, res);
			//select the maximum one
			if(curTriId != -1) //then res == 0, the point is on the triangle
			{			
				value = triToValue[curTriId];
				if(res > value)
				{
					triToValue[curTriId] = res;
					triToPoint[curTriId] = id;
				}
			}
			pointMap[id] = curTriId;
		}
		id += total;
	}
}
__global__ void cudaExpandTriangles(Triangle * triangles, int * pointMap, int * triToPoint, 
									REAL * triToValue, int * prefix, int oldTriNo)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	int pid, offset;
	Triangle * thisTri, * tri20, * tri01;
	
	while(id < oldTriNo)
	{		
		pid = triToPoint[id];
		if( pid>= 0)
		{			
			//basic info
			offset = prefix[id] + oldTriNo;
			thisTri = &triangles[id];

			//set pointMap
			pointMap[thisTri->vtx[0]] = offset;

			//replace
			//new triangle with 2-0(oppo 1) and with 0-1(oppo 2), The center point will be 0-index
			tri20 = &triangles[offset];
			tri01 = &triangles[offset + 1];
			tri20->vtx[0] = pid; tri20->vtx[1] = thisTri->vtx[2]; tri20->vtx[2] = thisTri->vtx[0];
			tri20->neighbor[0] = thisTri->neighbor[1]; tri20->neighbor[1]= Encode(offset+1, 2); tri20->neighbor[2] = Encode(id, 1);
			tri01->vtx[0] = pid; tri01->vtx[1] = thisTri->vtx[0]; tri01->vtx[2] = thisTri->vtx[1];
			tri01->neighbor[0] = thisTri->neighbor[2]; tri01->neighbor[1]= Encode(id, 2); tri01->neighbor[2] = Encode(offset, 1);

			thisTri->vtx[0] = pid; thisTri->neighbor[1] = Encode(offset, 2); thisTri->neighbor[2] = Encode(offset+1, 1);

			//update triToPoint, triToValue and pointToTri
			triToPoint[id] = -1; triToPoint[offset] = -1; triToPoint[offset + 1] = -1; 
			triToValue[id] = MINREAL; triToValue[offset] = MINREAL; triToValue[offset + 1] = MINREAL;

			prefix[id] = -1; //for next fix boundary
		}
		id += total;
	}
}


__global__ void cudaFixBoundary(Triangle * triangles, int * prefix, int oldNoTri, int newNoTriPair)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	Triangle * thisTri, * oppoTri;
	int oppoTriId, oppoPos;
	
	while(id < newNoTriPair)
	{
		int tri1 = oldNoTri + id * 2;
		int tri2 = tri1 + 1;

		//fix the boundary of tri1. Two case, seperate or not
		thisTri = & triangles[tri1];
		oppoTriId = Tri(thisTri->neighbor[0]);
		oppoPos = Pos(thisTri->neighbor[0]);
		oppoTri = & triangles[oppoTriId];
		if(prefix[oppoTriId] != -1)//not seperate
			oppoTri->neighbor[oppoPos] = Encode(tri1, 0);
		else if(thisTri->vtx[1] == oppoTri->vtx[1])	//seperate and case 1
			thisTri->neighbor[0] = Encode(Tri(oppoTri->neighbor[2]), 0);
		else if(thisTri->vtx[2] == oppoTri->vtx[2])	//seperate and case 1
			thisTri->neighbor[0] = Encode(Tri(oppoTri->neighbor[1]), 0);
		else //seperate and case 3
			oppoTri->neighbor[0] = Encode(tri1, 0);

		//fix the boundary of tri2
		thisTri = & triangles[tri2];
		oppoTriId = Tri(thisTri->neighbor[0]);
		oppoPos = Pos(thisTri->neighbor[0]);
		oppoTri = & triangles[oppoTriId];
		if(prefix[oppoTriId] != -1)//not seperate
			oppoTri->neighbor[oppoPos] = Encode(tri2, 0);
		else if(thisTri->vtx[1] == oppoTri->vtx[1])	//seperate and case 1
			thisTri->neighbor[0] = Encode(Tri(oppoTri->neighbor[2]), 0);
		else if(thisTri->vtx[2] == oppoTri->vtx[2])	//seperate and case 1
			thisTri->neighbor[0] = Encode(Tri(oppoTri->neighbor[1]), 0);
		else //seperate and case 3
			oppoTri->neighbor[0] = Encode(tri2, 0);

		id += total;
	}
}


__global__ void cudaSubdividePoints(Triangle * triangles, Point3D * points, int * pointMap, int * pointIndex,
		int * triToPoint, REAL * triToValue, Point3D kernel, int aliveNum)
{
	
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;

	int map, code, index, offset;
	Point3D ps[3], pc, pIndex;
	int pi[3];
	REAL res;

	while(id < aliveNum)
	{
		index = pointIndex[id];
		map = pointMap[index];
		Triangle thisTri =  triangles[map];
		if(index != thisTri.vtx[0] && index != thisTri.vtx[1] && index != thisTri.vtx[2]) //necessary for new assigned point
		{			
			pi[0] = triangles[Tri(thisTri.neighbor[1])].vtx[Pos(thisTri.neighbor[1])];
			pi[1] = thisTri.vtx[1];
			pi[2] = thisTri.vtx[2];
			ps[0] = points[pi[0]];
			ps[1] = points[pi[1]];
			ps[2] = points[pi[2]];
			pIndex = points[index];
			pc = points[thisTri.vtx[0]];

			res = g_check3DOrientation(kernel, pc, ps[0], pIndex, -1, thisTri.vtx[0], pi[0], index);
			if(res != 0.0)
			{
				code = res > 0? 1: 2;
				res = g_check3DOrientation(kernel, pc, ps[code], pIndex, -1, thisTri.vtx[0], pi[code], index);
				if(code == 1) offset = (res <= 0.0)?2: 0;
				else offset = (res >= 0.0)?1: 0;
			}
			else
			{
				res = g_check3DOrientation(kernel, ps[1], ps[2], pc, -1, pi[1], pi[2], thisTri.vtx[0]);
				if(res != 0.0) //offset = 0;
				{								
					//test c2i
					res = g_check3DOrientation(kernel, pc, ps[2], pIndex, -1, thisTri.vtx[0], pi[2], index); 
					if(res != 0.0) offset = (res<0)?0:1;
					else
					{
						//test c1i
						res = g_check3DOrientation(kernel, pc, ps[1], pIndex, -1, thisTri.vtx[0], pi[1], index); 
						offset = (res>=0)?0:2;
					}
				}
				else
				{
					res = g_check3DOrientation(kernel, ps[0], ps[2], pc, -1, pi[0], pi[2], thisTri.vtx[0]);
					if(res == 0.0) offset = 2;
					else offset = 1;
				}
			}
			
			//check whether is dead
			ps[offset] = pc;
			res = g_check3DOrientation( ps[0], ps[1], ps[2], pIndex, pi[0], pi[1], pi[2], index);
			if(res <= 0) {
				map = -1; //it is inside the new tetrahedra
			}
			else //record the furthest point
			{
				if(offset > 0) map = Tri(thisTri.neighbor[offset]);
				if(res >= triToValue[map])
				{
					triToValue[map] = res;
					triToPoint[map] = index;
				}
			}
			pointMap[index] = map;
		}

		id += total;
	}
}
__global__ void cudaLabelAlivePoints(Triangle * triangles, int * pointMap, int * pointIndex, bool * isAliveArray, int noAlive)
{
	
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	int map, index;
	bool isAlive;
	Triangle * tri;

	while(id < noAlive)
	{
		index = pointIndex[id];
		map = pointMap[index];
		isAlive = true;
		if(map == -1)
			isAlive = false;
		else
		{
			tri = & triangles[map];
			if(index == tri->vtx[0] || index == tri->vtx[1] || index == tri->vtx[2])
				isAlive = false;
		}
		isAliveArray[id] = isAlive;
		id += total;
	}
}
__device__ void cd_LockTriOfNonExtreme(int * triLabels, Triangle * triangles, int triId, 
									   int leftTriId, int rightTriId, int leftPos)
{
	int nextCode;
	atomicMin(triLabels + leftTriId, -triId); //atomMin triangle label, non extreme have highest priority
	leftPos = Prev(leftPos);
	while(leftTriId != rightTriId)
	{									
		nextCode = triangles[leftTriId].neighbor[leftPos];
		leftTriId = Tri(nextCode);
		leftPos = Prev(Pos(nextCode));
		atomicMin(triLabels + leftTriId, -triId);
	}
}
__global__ void cudaCheckTriangles(int * triIndex, int * triStatus, int * triLabels, Triangle * triangles, 
								   Point3D * points, int * pointMap, Point3D kernel, int noTri, int checkNum)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	int triId, pos, oppoTriId, oppoPos;
	int status, code, offset;
	int pid, oppoPid, nextPid, prevPid;
	Triangle tri, *oppoTri;
	Point3D ps[3];
	unsigned int edgeOffset;
	REAL edgeRes, conRes;
	bool toFlip;

	while(id < checkNum)
	{
		triId = triIndex[id];
		status = triStatus[triId];
		tri = triangles[triId];
		ps[0] = points[tri.vtx[0]];
		ps[1] = points[tri.vtx[1]];
		ps[2] = points[tri.vtx[2]];

		//check the first edge
		for(pos = 0, edgeOffset = 1; pos < 3; pos ++, edgeOffset <<= 1)
		{
			if((status & edgeOffset) == 0) continue;
			oppoTriId = Tri(tri.neighbor[pos]);
			oppoPos = Pos(tri.neighbor[pos]);
			//if this triId is larger and the opposite one also check this, set it be 0 and continue
			if(triId > oppoTriId && (triStatus[oppoTriId] &(1<<oppoPos))!= 0) 
			{
				status = status & (~edgeOffset);
				continue;
			}
			//begin to check the edge
			code = 0;
			oppoTri = &triangles[oppoTriId];
			pid = tri.vtx[pos];
			oppoPid = oppoTri->vtx[oppoPos];
			nextPid = tri.vtx[Next(pos)];
			prevPid = tri.vtx[Prev(pos)];
			edgeRes = g_check3DOrientation(ps[pos], ps[Next(pos)], ps[Prev(pos)], points[oppoPid],
				tri.vtx[pos], tri.vtx[Next(pos)], tri.vtx[Prev(pos)], oppoPid); //<0: convex

			//if it is reflex, continue
			//check the topology.Because we keep star shape, so any 3-1 topology can be flippable.
			if(Tri(tri.neighbor[Prev(pos)]) == Tri(oppoTri->neighbor[Next(oppoPos)])) //3-1, and pidNext is in the cone
			{
				//in gpu, we donot remove coplanar point, for efficiency
				if(edgeRes > 0 || pointMap[nextPid] == -1) //edge is reflex or pidNext degree is 3
				{
					code = 4;//100;
					atomicMin(triLabels + triId, triId); //atomMin triangle label, 3-1 flipping have middle priority
					atomicMin(triLabels + oppoTriId, triId);
					atomicMin(triLabels + Tri(tri.neighbor[Prev(pos)]), triId);
				}
			}
			else if(Tri(tri.neighbor[Next(pos)]) == Tri(oppoTri->neighbor[Prev(oppoPos)])) //3-1, and pidPrev is in the cone
			{
				if(edgeRes > 0 || pointMap[prevPid] == -1) // edge is reflex or pidPrev degree is 3
				{
					code = 5;//101;
					atomicMin(triLabels + triId, triId); //atomMin triangle label, 3-1 flipping have middle priority
					atomicMin(triLabels + oppoTriId, triId);
					atomicMin(triLabels + Tri(tri.neighbor[Next(pos)]), triId);
				}
			}
			else //if the topology is 2-2
			{
				//conRes = g_check3DOrientation(points, kernel, -1, pid, oppoPid, nextPid);
				conRes = g_check3DOrientation(kernel, ps[pos], points[oppoPid], ps[Next(pos)], 
					-1, tri.vtx[pos], oppoPid, tri.vtx[Next(pos)]);
				if(conRes >= 0 && edgeRes > 0 && pointMap[nextPid]!=-1) 
				// pidNext is inside the cone kernel-pid1-pid2-pidPrev + reflex ->unflippable
				{
					//label pidNext
					code = 2;//010; 
					pointMap[nextPid] = -1;
					cd_LockTriOfNonExtreme(triLabels, triangles, triId, triId, oppoTriId, pos);
				}
				else if(conRes <= 0)
				{
					conRes = g_check3DOrientation(kernel, points[oppoPid], ps[pos], ps[Prev(pos)],
						-1, oppoPid, tri.vtx[pos], tri.vtx[Prev(pos)]);
					if(conRes >= 0 && edgeRes > 0 && pointMap[prevPid]!=-1) 
					//pidPrev is inside the conve kernel-pid2-pid1-pidNext + reflex -> unflippable
					{
						//label pidPrev
						code = 3;//011; 
						pointMap[prevPid] = -1;
						cd_LockTriOfNonExtreme(triLabels, triangles, triId, oppoTriId, triId, oppoPos);
					}
					else if(conRes <= 0) //flippable
					{
						offset = MAXINT;
						if(pointMap[pid] == -1) offset = __min(offset, pid);
						if(pointMap[oppoPid] == -1) offset = __min(offset, oppoPid);
						if(pointMap[nextPid] == -1) offset = __min(offset, nextPid);
						if(pointMap[prevPid] == -1) offset = __min(offset, prevPid);
						if(offset == MAXINT) toFlip = (edgeRes > 0);
						else toFlip = (offset == nextPid)|(offset == prevPid);

						if(toFlip)
						{
							code = 6;//110; 
							atomicMin(triLabels + triId, triId + noTri); //atomMin triangle label, 2-2 flipping have lowest priority
							atomicMin(triLabels + oppoTriId, triId + noTri);
						}
					}
				}
			}
			//check the code
			//if the code is still 0, label the edge bit 0 and continue;
			if(code == 0)
			{
				status = status & (~edgeOffset);
			}
			//else, set the code in corresponding status bits. Then atomicMin the triangle, move up
			else
			{
				offset = (pos+1)*3;
				status = ((status>>(offset +3))<<(offset+3)) + (code<<offset) + (status & ((1<<offset)-1));
				break;
			}
		}
		triStatus[triId] = status;

		id += total;
	}
}
__device__ void cd_clearTriangle(Triangle * tri)
{
	tri->vtx[0]/* = tri->vtx[1] = tri->vtx[2] = tri->neighbor[0] = tri->neighbor[1] = tri->neighbor[2] */= -1;
}
__global__ void cudaPrepareFlipping(int * triIndex, int * triStatus, int * triLabels, Triangle * triangles, 
									int noTri, int checkNum)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	int triId, oppoTriId, pos, label, oppoLabel, sharedTriId, sharedLabel, status, code;
	Triangle * tri;

	while(id < checkNum)
	{
		triId = triIndex[id];
		status = triStatus[triId];
		tri = & triangles[triId];
		label = triLabels[triId];

		//check the first edge
		pos = 0;
		while(((status>>pos)&1)==0 && pos<3) pos ++;
		if(pos < 3) //means some edge needs to be checked
		{
			oppoTriId = Tri(tri->neighbor[pos]);
			oppoLabel = triLabels[oppoTriId];
			code = (status>>((pos+1) *3)) & 7;
			
			if(code == 2 || code == 3)
			{
				//non extreme point
				status = -status;
			}
			else if(code == 4 || code == 5)
			{
				//3-1
				if(code == 4)
					sharedTriId = Tri(tri->neighbor[Prev(pos)]);
				else
					sharedTriId = Tri(tri->neighbor[Next(pos)]);
				sharedLabel = triLabels[sharedTriId];
				if(label == triId && oppoLabel  == triId && sharedLabel == triId)
				{
					status = -status;
				}
				else //reset label which equal to me
				{
					if(label == triId)
					{
						triLabels[triId] = MAXINT;
					}
					if(oppoLabel == triId)
					{
						triLabels[oppoTriId] = MAXINT;
					}
					if(sharedLabel == triId)
					{
						triLabels[sharedTriId] = MAXINT;
					}
				}
			}
			else if(code == 6)
			{
				//2-2 flipping
				if(label == triId + noTri && oppoLabel  == triId + noTri)
				{
					status = -status;
				}
				else //reset label which equal to me
				{
					if(label == triId + noTri)
					{
						triLabels[triId] = MAXINT;
					}
					if(oppoLabel == triId + noTri)
					{
						triLabels[oppoTriId] = MAXINT;
					}
				}
			}
			triStatus[triId] = status;
		}
		id += total;
	}
}
__device__ void cd_LabelTriOfNonExtreme(int * triStatus, int * triLabels, Triangle * triangles, int triId,
										int leftTriId, int rightTriId, int leftPos)
{
	int nextCode;
	//make all the label to maxint, make all except left and right to all check
	if(leftTriId != triId && triStatus[leftTriId] >=0)
		triLabels[leftTriId] = MAXINT; //last 3 bits: 00(inner pos to tri1)001(3-1 flipping and kept)
	leftPos = Prev(leftPos);
	while(leftTriId != rightTriId)
	{						
		nextCode = triangles[leftTriId].neighbor[leftPos];
		leftTriId = Tri(nextCode);
		leftPos = Prev(Pos(nextCode));
		if(leftTriId != triId && triStatus[leftTriId] >=0) 
			triLabels[leftTriId] = MAXINT; 
	}
}
__device__ void cd_flip31(int * triStatus, int * triLabels, Triangle * triangles, int * pointMap, 
						  int triId1, int pos1, int triId2, int pos2)//1-left; 2-right. left is source, always
{
	Triangle * tri1 = &triangles[triId1];
	Triangle * tri2 = &triangles[triId2];
	int sharedTriId = Tri(tri1->neighbor[Prev(pos1)]);
	int sharedPos = Pos(tri1->neighbor[Prev(pos1)]);
	Triangle * sharedTri = & triangles[sharedTriId];
	
	int pid = tri1->vtx[Next(pos1)];
	int pid1 = tri1->vtx[pos1];
	int pid2 = tri2->vtx[pos2];
	int pidOut = tri1->vtx[Prev(pos1)];

	//remove triId1, triId2 and sharedTriId
	//put the new triangle on triId1
	tri1->neighbor[pos1] = tri2->neighbor[Prev(pos2)];
	tri1->neighbor[Prev(pos1)] = sharedTri->neighbor[Next(sharedPos)];
	tri1->vtx[Next(pos1)] = pid2;

	//"remove" tri2 and sharedTri
	cd_clearTriangle(tri2); //may be triId, consider it?
	cd_clearTriangle(sharedTri);

	//update triStatus for next rounding checking
	triStatus[triId1] = 0x00000007;
	triStatus[triId2] = 0x00000000;	//never use
	triStatus[sharedTriId] = 0x00000000; //never use

	//code triLabels[triId1], triLabels[triId2] and triLabels[sharedTri] for topology repair
	triLabels[triId1] = (((Next(pos1))<<3) | 1) | 0x80000000;	//last 3 bits: 00(inner pos to tri1)001(3-1 flipping and kept)
	triLabels[triId2] = ((pos2<<3) | 2) | 0x80000000;	//last 3 bits: 00(oppoPos to tri1)010(3-1 flipping and removed)
	triLabels[sharedTriId] = ((sharedPos<<3) | 2) | 0x80000000;	//last 3 bits: 00(oppoPos to tri1)010(3-1 flipping and removed)

	//update pointMap
	pointMap[pid] = -1;
	if(pointMap[pid1]>=0)
		pointMap[pid1] = triId1;
	if(pointMap[pid2]>=0)
		pointMap[pid2] = triId1;
	if(pointMap[pidOut]>=0)
		pointMap[pidOut] = triId1;
}
__device__ void cd_flip22(int * triStatus, int * triLabels, Triangle * triangles, int * pointMap, 
						  int triId1, int pos1, int triId2, int pos2, int noTri)
{	
	Triangle * tri1 = &triangles[triId1];
	Triangle * tri2 = &triangles[triId2];
	
	int pid1 = tri1->vtx[pos1];
	int pid2 = tri2->vtx[pos2];
	int pidPrev = tri1->vtx[Prev(pos1)];
	int pidNext = tri1->vtx[Next(pos1)];

	//update tri1 and tri2
	tri1->vtx[Next(pos1)] = pid2;
	tri2->vtx[Next(pos2)] = pid1;
	tri1->neighbor[pos1] = tri2->neighbor[Prev(pos2)];
	tri2->neighbor[pos2] = tri1->neighbor[Prev(pos1)];
	tri1->neighbor[Prev(pos1)] = Encode(triId2, Prev(pos2));
	tri2->neighbor[Prev(pos2)] = Encode(triId1, Prev(pos1));

	//update triStatus for next rounding checking
	triStatus[triId1] = (1<<pos1) + (1<<(Next(pos1)));
	triStatus[triId2] = (1<<pos2) + (1<<(Next(pos2)));
	
	//code triLabels[triId1], triLabels[triId2] for topology repair
	triLabels[triId1] = (pos1<<3) | 0x80000000;	//last 3 bits: 00(pos1)000(2-2 flipping, and source)
	triLabels[triId2] = ((pos2<<3) | 3) | 0x80000000;	//last 3 bits: 00(pos1)011(2-2 flipping, and destination)

	//update pointMap
	if(pointMap[pidPrev]>=0)
		pointMap[pidPrev] = triId1;
	if(pointMap[pidNext]>=0)
		pointMap[pidNext] = triId2;
}
		
__global__ void cudaFlipTriangles(int * triIndex, int * triStatus, int * triLabels, Triangle * triangles, 
								  int * pointMap, int noTri, int checkNum)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	int triId, pos, oppoTriId, oppoPos;
	int status, code;
	int leftTriId, leftPos, rightTriId, rightPos;
	Triangle * tri;

	while(id < checkNum)
	{
		triId = triIndex[id];
		status = triStatus[triId];
		tri = & triangles[triId];

		if(status < 0) //means the triangle is active triangle, all induced sub complex belong to it
		{
			status = - status;
			pos = 0;
			while(((status>>pos)&1)==0 && pos<3) pos ++; //find the pos
			oppoTriId = Tri(tri->neighbor[pos]);
			oppoPos = Pos(tri->neighbor[pos]);
			code = (status>>((pos+1) *3)) & 7;

			if(code == 2 || code == 3)
			{
				//set the surrounding triangle to be checked
				if(code == 2)
				{
					leftTriId = triId;
					leftPos = pos;
					rightTriId = oppoTriId;
				}
				else
				{
					leftTriId = oppoTriId;
					leftPos = oppoPos;
					rightTriId = triId;
				}
				cd_LabelTriOfNonExtreme(triStatus, triLabels, triangles, triId, leftTriId, rightTriId, leftPos);
				//use Label[triId] to record
				triLabels[triId] = ((pos<<3) | (code<<1)) | 0x80000000;	//last 3 bits: 00(pos1)100/110(2-2 flipping, and source)
			}
			else if(code == 4 || code == 5)
			{
				//3-1 flipping, must keep left triangle to be triId
				if(code == 4)
				{
					leftTriId = triId;
					leftPos = pos;
					rightTriId = oppoTriId;
					rightPos = oppoPos;
				}
				else
				{
					leftTriId = triId;
					leftPos = Next(pos);
					rightTriId = Tri(tri->neighbor[leftPos]);
					rightPos = Pos(tri->neighbor[leftPos]);
				}
				cd_flip31(triStatus, triLabels, triangles, pointMap, leftTriId, leftPos, rightTriId, rightPos);
			}
			else if(code == 6)
			{
				//2-2 flipping
				cd_flip22(triStatus, triLabels, triangles, pointMap, triId, pos, oppoTriId, oppoPos, noTri);
			}
		}

		id += total;
	}
}


__device__ unsigned int cd_correctTriCode(int * triLabels, Triangle * triangles, int triId, int triPos, bool & staticTri)
{
	//some triangle launch the signal triId-pos
	//return the new code and tell whether this is a static triangle
	int code = 0;
	staticTri = true; //suppose it is a static triangle
	int label = triLabels[triId];
	int type, pos;
	if(label < 0) //means special coding
	{
		staticTri = false;
		type = label & 7;
		pos = (label>>3) & 3;
		if(type == 0 || type == 3) //00/11: 2-2 flipping source | dest
		{
			if(triPos == Prev(pos))
			{
				code = triangles[triId].neighbor[triPos];
				code = Encode(Tri(code), Next(Pos(code)));
			}
			else if(triPos == Next(pos))
			{
				code = Encode(triId, triPos);
			}
		}
		else if(type == 1) //01: 3-1 flipping kept
		{
			code = Encode(triId, triPos);			
		}
		else if(type == 2) //10: 3-1 flipping remove
		{
			code = triangles[triId].neighbor[pos];
		}
		else if(type == 4 || type == 6)
		{
			//non extreme case, the triangle does not change
			staticTri = true;
		}
	}
	return code;
}
__device__ void cd_correctStatusOfNonExtreme(int * triStatus, Triangle * triangles, int leftTriId, int rightTriId, int leftPos)
{
	int nextCode;
	leftPos = Prev(leftPos);
	while(leftTriId != rightTriId)
	{						
		nextCode = triangles[leftTriId].neighbor[leftPos];
		leftTriId = Tri(nextCode);
		leftPos = Pos(nextCode);
		if(leftTriId != rightTriId) triStatus[leftTriId] = 7;//add one check on one edge is enough
		leftPos = Prev(leftPos);//for next round
	}
}
__global__ void cudaFixTopology(int * triIndex, int * triStatus, int * triLabels, Triangle * triangles, int checkNum)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	int label, type, code;
	int triId, pos, oppoTriId, oppoPos;
	int leftTriId, rightTriId, leftPos;
	Triangle * tri;
	bool staticTri;
	
	while(id < checkNum)
	{
		triId = triIndex[id];
		label = triLabels[triId];
		tri = &triangles[triId];
	
		if(label < 0) //means special coding 
		{
			type = label & 7;
			pos = (label>>3) & 3;
			if(type == 0)//2-2 flipping source
			{
				//work:1. repair the correct neighbor of two. 2. fix the neighbor for boundary if those triLabels=MAXINT
				code = tri->neighbor[pos];
				oppoTriId = Tri(code);
				oppoPos = Pos(code);
				code = cd_correctTriCode(triLabels, triangles, oppoTriId, oppoPos, staticTri);
				if(!staticTri) tri->neighbor[pos] = code;
				else triangles[oppoTriId].neighbor[oppoPos] = Encode(triId, pos);

				pos = Next(pos);
				code = tri->neighbor[pos];
				oppoTriId = Tri(code);
				oppoPos = Pos(code);
				code = cd_correctTriCode(triLabels, triangles, oppoTriId, oppoPos, staticTri);
				if(!staticTri) tri->neighbor[pos] = code;
				//else do not need to update

				//move to next triangle
				code = tri->neighbor[Next(pos)];
				triId = Tri(code);
				pos = Next(Pos(code));
				tri = &triangles[triId];
				
				code = tri->neighbor[pos];
				oppoTriId = Tri(code);
				oppoPos = Pos(code);
				code = cd_correctTriCode(triLabels, triangles, oppoTriId, oppoPos, staticTri);
				if(!staticTri) tri->neighbor[pos] = code;
				else triangles[oppoTriId].neighbor[oppoPos] = Encode(triId, pos);
				
				pos = Next(pos);
				code = tri->neighbor[pos];
				oppoTriId = Tri(code);
				oppoPos = Pos(code);
				code = cd_correctTriCode(triLabels, triangles, oppoTriId, oppoPos, staticTri);
				if(!staticTri) tri->neighbor[pos] = code;
				//else do not need to update
			}
			else if(type == 1)//3-1 flipping source
			{
				//judge whether it need work
				code = tri->neighbor[pos];
				oppoTriId = Tri(code);
				oppoPos = Pos(code);
				code = cd_correctTriCode(triLabels, triangles, oppoTriId, oppoPos, staticTri);
				if(!staticTri) tri->neighbor[pos] = code;
				//else do not need to update
				
				//move to the next edge
				pos = Next(pos);
				code = tri->neighbor[pos];
				oppoTriId = Tri(code);
				oppoPos = Pos(code);
				code = cd_correctTriCode(triLabels, triangles, oppoTriId, oppoPos, staticTri);
				if(!staticTri) tri->neighbor[pos] = code;
				else triangles[oppoTriId].neighbor[oppoPos] = Encode(triId, pos);
				
				//move to the next edge
				pos = Next(pos);
				code = tri->neighbor[pos];
				oppoTriId = Tri(code);
				oppoPos = Pos(code);
				code = cd_correctTriCode(triLabels, triangles, oppoTriId, oppoPos, staticTri);
				if(!staticTri) tri->neighbor[pos] = code;
				else triangles[oppoTriId].neighbor[oppoPos] = Encode(triId, pos);
			}
			else if(type == 4 || type == 6)
			{
				code = tri->neighbor[pos];
				oppoTriId = Tri(code);
				oppoPos = Pos(code);
				//go around and label all status to |7 or ...
				if(type == 4)
				{
					leftTriId = triId;
					leftPos = pos;
					rightTriId = oppoTriId;
				}
				else
				{
					leftTriId = oppoTriId;
					leftPos = oppoPos;
					rightTriId = triId;
				}
				cd_correctStatusOfNonExtreme(triStatus, triangles, leftTriId, rightTriId, leftPos);
			}
		}
		id += total;
	}
}
__global__ void cudaClearLabel(int * triIndex, int * triStatus, int* triLabels, Triangle * triangles, 
							   bool * g_check, int checkNum)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x;
	int total = gridDim.x * blockDim.x;
	bool toCheck;
	int triId, label, type, pos, status;
	Triangle * tri;

	while(id < checkNum)
	{
		toCheck = false;
		triId = triIndex[id];
		label = triLabels[triId];
		status = triStatus[triId];
		if(label < 0) //means special coding 
		{
			if(status < 0) triStatus[triId] = - status;
			type = label & 7;
			pos = (label>>3) & 3;
			if(type == 0)//2-2 flipping source
			{
				//clear
				tri = &triangles[triId];
				triLabels[triId] = MAXINT;
				triId = Tri(tri->neighbor[Prev(pos)]);
				triLabels[triId] = MAXINT;
			}
			else //type = 1(3-1) 4|6(non extreme)
			{
				//clear
				triLabels[triId] = MAXINT;
				//the removed one does not need to remove, because they are not on the polytope
			}

			toCheck = true;
		}
		if(toCheck) g_check[0] = true;
		id += total;
	}
}

__global__ void cudaCompactTriangles(Triangle *oldTriangles, Triangle * newTriangles, int * prefix, int oldNoTri)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;

	while(id < oldNoTri)
	{
		int newIndex = prefix[id];
		Triangle tri = oldTriangles[id];
		if(tri.vtx[0] != -1)
		{
			tri.neighbor[0] = Encode(prefix[Tri(tri.neighbor[0])], Pos(tri.neighbor[0]));
			tri.neighbor[1] = Encode(prefix[Tri(tri.neighbor[1])], Pos(tri.neighbor[1]));
			tri.neighbor[2] = Encode(prefix[Tri(tri.neighbor[2])], Pos(tri.neighbor[2]));
			newTriangles[newIndex] = tri;
		}

		id += total;
	}
}

__global__ void cudaUpdatePointMap(int * pointNum, int * spaces, int noPoint)
{
	int id = blockIdx.x * blockDim.x + threadIdx.x; 
	int total = gridDim.x * blockDim.x;
	int oldTriId;

	while(id < noPoint)
	{
		oldTriId = pointNum[id];
		if(oldTriId >=0)
			pointNum[id] = spaces[oldTriId];
		id += total;
	}
}