/*
Author: Gao Mingcen
Date: 28/02/2013

File Name: GLFunction.h

Functions for visualization in OpenGL

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#pragma once

#include "glew.h"
#include "glut.h"
#include "TrackBall.h"
#include <ctype.h>

TriangleMesh3D * vizMesh;
TrackBall trackball;
bool showPoints = true; 
bool showFrames = true;
bool showTriangles = true; 
bool showBox = true;


void normalVector(Point3D * v0, Point3D *v1, Point3D *v2, REAL *n)
{
    float a[3], b[3]; 

    a[0] = v0->x - v1->x; 
    a[1] = v0->y - v1->y; 
    a[2] = v0->z - v1->z; 

    b[0] = v1->x - v2->x; 
    b[1] = v1->y - v2->y; 
    b[2] = v1->z - v2->z; 

    n[0] = (a[1] * b[2]) - (a[2] * b[1]);
    n[1] = (a[2] * b[0]) - (a[0] * b[2]);
    n[2] = (a[0] * b[1]) - (a[1] * b[0]);

    // Normalize
    float dist = sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]); 

    if (dist == 0.0) 
        dist = 1.0; 

    n[0] /= -dist; 
    n[1] /= -dist; 
    n[2] /= -dist; 
}

void drawBox()
{
    glColor3d(0.5, 0.5, 0.5); 

    glBegin(GL_LINE_LOOP); 
        glVertex3d(0.0, 0.0, 0.0); glVertex3d(0.0, 1.0, 0.0); 
        glVertex3d(1.0, 1.0, 0.0); glVertex3d(1.0, 0.0, 0.0); 
    glEnd(); 
    glBegin(GL_LINE_LOOP); 
        glVertex3d(0.0, 0.0, 1.0); glVertex3d(0.0, 1.0, 1.0); 
        glVertex3d(1.0, 1.0, 1.0); glVertex3d(1.0, 0.0, 1.0); 
    glEnd(); 
    glBegin(GL_LINES); 
        glVertex3d(0.0, 0.0, 0.0); glVertex3d(0.0, 0.0, 1.0);
        glVertex3d(0.0, 1.0, 0.0); glVertex3d(0.0, 1.0, 1.0);
        glVertex3d(1.0, 1.0, 0.0); glVertex3d(1.0, 1.0, 1.0);
        glVertex3d(1.0, 0.0, 0.0); glVertex3d(1.0, 0.0, 1.0);
    glEnd(); 
}

void drawPoints()
{
	glPointSize(2.0); 
	glColor3f(1.0, 1.0, 0.0);

    glBegin(GL_POINTS); 
	for(int i = 0; i < vizMesh->noPoint; i++)
	{
		Point3D * point =  & vizMesh->points[i];
		glVertex3d(double(point->x), double(point->y), double(point->z)); 
	}
	glEnd();
}

void drawFrames()
{
	glColor3d(1.0, 0.3, 0.3); 
	//draw line
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); 
    glBegin(GL_LINES);

	for(int tid = 0; tid < vizMesh->noTri; tid++)
	{
		Triangle * tri = & vizMesh->triangles[tid];
		if(tri->vtx[0] == -1) 
			continue;
		for(int pos = 0; pos < 3; pos++)
		{
			int pid1 = tri->vtx[pos];
			int pid2 = tri->vtx[Next(pos)];
			int neiCode = tri->neighbor[Prev(pos)];
			if(neiCode == -1) continue;
			int neiTid = Tri(neiCode);
			if(tid < neiTid)
			{
				Point3D * point1 =  & vizMesh->points[pid1];
				Point3D * point2 =  & vizMesh->points[pid2];
				glVertex3d(double(point1->x), double(point1->y), double(point1->z)); 
				glVertex3d(double(point2->x), double(point2->y), double(point2->z)); 
			}
		}
	}
	glEnd();
}

void drawTriangles()
{	
	glEnable (GL_BLEND); 
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glFrontFace(GL_CCW);
    //glPolygonMode(GL_FRONT, GL_FILL); 
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL); 
	glEnable(GL_LIGHTING); 
	glEnable(GL_LIGHT0); 
	glPushMatrix();
	glMatrixMode(GL_MODELVIEW); 
	glLoadIdentity();
	GLfloat ambient[] = { .0f, .0f, .0f }; 
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient); 
	GLfloat difuse[] = { .9f, .9f, .9f }; 
	glLightfv(GL_LIGHT0, GL_DIFFUSE, difuse); 
	GLfloat position[] = { 4.0 , -4.0 , 0.0 }; 
	glLightfv(GL_LIGHT0, GL_POSITION, position); 
	glPopMatrix(); 

	float mcolor[] = { .9f, .9f, .9f, .9f}; 
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mcolor);
    glShadeModel(GL_FLAT); 
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0, 1.0);

	glBegin(GL_TRIANGLES); 		
	REAL normal[3]; 

	for(int tid = 0; tid < vizMesh->noTri; tid ++)
	{
		Triangle * tri =  & vizMesh->triangles[tid];
		if(tri->vtx[0] == -1) continue;
		Point3D * point0 = &vizMesh->points[tri->vtx[0]];
		Point3D * point1 = &vizMesh->points[tri->vtx[1]];
		Point3D * point2 = &vizMesh->points[tri->vtx[2]];
		normalVector(point0, point1, point2, normal); 
		glNormal3d(normal[0], normal[1], normal[2]); 
		glVertex3d(double(point0->x), double(point0->y), double(point0->z)); 
		glVertex3d(double(point1->x), double(point1->y), double(point1->z)); 
		glVertex3d(double(point2->x), double(point2->y), double(point2->z)); 

	}

	glEnd();
	
    glDisable(GL_LIGHT0); 
    glDisable(GL_LIGHTING); 
    glDisable(GL_POLYGON_OFFSET_FILL);
    glPolygonMode(GL_FRONT, GL_FILL);
}

void gl_display()
{
	glViewport(0,0,512,512);
	glClearColor(0,0,0,0); 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
	// Setup projection matrix
	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity(); 
	glOrtho(0-0.1, 1+0.1, 0-0.1, 1+0.1, -1.0*100, 1.0*100);

	// Setup modelview matrix
	glMatrixMode(GL_MODELVIEW); 
	glLoadIdentity();

    glTranslated(trackball.GetTransPtr()[0], trackball.GetTransPtr()[1], 0.0);   
	glTranslated(trackball.GetOffsetPtr()[0], trackball.GetOffsetPtr()[1], trackball.GetOffsetPtr()[2]);
    glScaled(*trackball.GetScalePtr(), *trackball.GetScalePtr(), *trackball.GetScalePtr()); 
    glMultMatrixf(trackball.GetRotationMatrix()); 
	glTranslated(-trackball.GetOffsetPtr()[0], -trackball.GetOffsetPtr()[1], -trackball.GetOffsetPtr()[2]);

	// Setting up lights
	glDisable(GL_LIGHTING); 
	glEnable(GL_DEPTH_TEST); 

	glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0, 1.0);

	//begin to Draw
	if(showBox) drawBox();
	if(showPoints) drawPoints();
	if(showFrames) drawFrames();
	if(showTriangles) drawTriangles();
	
	glutSwapBuffers(); 
}

void gl_reshape(int width, int height)
{
	float winSize = min( width, height ); 
	trackball.SetWindowSize(width, height);
	
	//glutSwapBuffers(); 
}

void gl_mouseMotion(int x, int y)
{
	trackball.ReceiveMouseMotionFunc(x,y);
}

void gl_keyboard(unsigned char key, int x, int y)
{
	switch (toupper(key))
	{
    case 'P':
        showPoints = !showPoints; 
        break; 
    case 'B':
        showBox = !showBox; 
        break; 
    case 'T':
        showTriangles = !showTriangles; 
        break; 
    case 'F':
        showFrames = !showFrames; 
        break; 
    case 'Q':
        trackball.Rotate(1.0, 0.0, 0.0, -3.141592654 / 100.0); 
        break; 
    case 'A':
        trackball.Rotate(1.0, 0.0, 0.0, 3.141592654 / 100.0); 
        break; 
    case 'W':
        trackball.Rotate(0.0, 1.0, 0.0, -3.141592654 / 100.0); 
        break; 
    case 'S':
        trackball.Rotate(0.0, 1.0, 0.0, 3.141592654 / 100.0); 
        break; 
    case 'E':
        trackball.Rotate(0.0, 0.0, 1.0, -3.141592654 / 100.0); 
        break; 
    case 'D':
        trackball.Rotate(0.0, 0.0, 1.0, 3.141592654 / 100.0); 
		break;
	case 'G':
		trackball.Reset();
		break;
	}
}

void gl_mouse(int button, int state, int x, int y)
{
	trackball.ReceiveMouseFunc(button, state, x, y);
}

void gl_visualize(TriangleMesh3D * thisMesh)
{
	vizMesh = thisMesh;
	int argc = 1; char *argv = "FFHull.exe";
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA|GLUT_STENCIL); 
	glutInit(&argc, &argv);  
	glutInitDisplayMode (GLUT_DOUBLE);
	glutInitWindowPosition(400, 0); 
	glutInitWindowSize(512, 512);
	glutCreateWindow("FFHull");
	glewInit();

	printf("------FFHull Visualization------\n");
	printf("*** P: show/hide points\n");
	printf("*** B: show/hide box\n");
	printf("*** F: show/hide frames\n");
	printf("*** T: show/hide triangles\n");
	printf("*** Q/A/W/S/E/D: rotate the scene\n");
	printf("*** G: reset the scene\n");
	printf("\n");

	glutReshapeFunc(gl_reshape);
	glutMouseFunc(gl_mouse); 
	glutMotionFunc(gl_mouseMotion); 
	glutKeyboardFunc(gl_keyboard);
	glutDisplayFunc(gl_display);
	glutIdleFunc(gl_display);
	glutMainLoop();
}