/*
Author: Cao Thanh Tung & Gao Mingcen
Date: 28/02/2013

File Name: TrackBall.cpp

Functions of TrackBall, a controller of 3D space for visualization

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <glut.h>
#include <math.h>
#include <algorithm>
#include "TrackBall.h"

#define PI      3.141592654

TrackBall::TrackBall()
{
    // Set default window size
    width = 512; height = 512; 
	leftButton = true;
	rightButton = true;
	middleButton = true;

    // Initialize the rotation matrix
	rotation = new float[16];
    rotation[0 * 4 + 0] = 1.0; 
    rotation[0 * 4 + 1] = 0.0; 
    rotation[0 * 4 + 2] = 0.0; 
    rotation[0 * 4 + 3] = 0.0; 
    rotation[1 * 4 + 0] = 0.0; 
    rotation[1 * 4 + 1] = 1.0; 
    rotation[1 * 4 + 2] = 0.0; 
    rotation[1 * 4 + 3] = 0.0; 
    rotation[2 * 4 + 0] = 0.0; 
    rotation[2 * 4 + 1] = 0.0; 
    rotation[2 * 4 + 2] = 1.0; 
    rotation[2 * 4 + 3] = 0.0; 
    rotation[3 * 4 + 0] = 0.0; 
    rotation[3 * 4 + 1] = 0.0; 
    rotation[3 * 4 + 2] = 0.0; 
    rotation[3 * 4 + 3] = 1.0; 

	//initialize the parameters
	
	isLeftMouseActive      = false;
    isRightMouseActive     = false; 
    isMiddleMouseActive    = false; 
    oldMouseX   = 0; 
    oldMouseY   = 0; 
    mouseMove = false;
	scale     = 1.0;
    trans[0]    = 0.0;
    trans[1]    = 0.0;
	this->offset[0] = this->offset[1]= this->offset[2]= 0.5;
	this->offsetRatio = 0.95;
    // Reset the rotation
    Reset(); 
}

TrackBall::TrackBall(int w, int h)
{
    // Set default window size
    width = w; height = h; 
	leftButton = true;
	rightButton = true;
	middleButton = true;

    // Initialize the rotation matrix
	rotation = new float[16];
    rotation[0 * 4 + 0] = 1.0; 
    rotation[0 * 4 + 1] = 0.0; 
    rotation[0 * 4 + 2] = 0.0; 
    rotation[0 * 4 + 3] = 0.0; 
    rotation[1 * 4 + 0] = 0.0; 
    rotation[1 * 4 + 1] = 1.0; 
    rotation[1 * 4 + 2] = 0.0; 
    rotation[1 * 4 + 3] = 0.0; 
    rotation[2 * 4 + 0] = 0.0; 
    rotation[2 * 4 + 1] = 0.0; 
    rotation[2 * 4 + 2] = 1.0; 
    rotation[2 * 4 + 3] = 0.0; 
    rotation[3 * 4 + 0] = 0.0; 
    rotation[3 * 4 + 1] = 0.0; 
    rotation[3 * 4 + 2] = 0.0; 
    rotation[3 * 4 + 3] = 1.0; 

	//initialize the parameters
	
	isLeftMouseActive      = false;
    isRightMouseActive     = false; 
    isMiddleMouseActive    = false; 
    oldMouseX   = 0; 
    oldMouseY   = 0; 
    mouseMove = false;
	scale     = 1.0;
    trans[0]    = 0.0;
    trans[1]    = 0.0;

    // Reset the rotation
    Reset(); 
}

TrackBall::TrackBall(int w, int h, bool left, bool right, bool middle)
{
    // Set default window size
    width = w; height = h;  
	leftButton = left;
	rightButton = right;
	middleButton = middle;

    // Initialize the rotation matrix
	rotation = new float[16];
    rotation[0 * 4 + 0] = 1.0; 
    rotation[0 * 4 + 1] = 0.0; 
    rotation[0 * 4 + 2] = 0.0; 
    rotation[0 * 4 + 3] = 0.0; 
    rotation[1 * 4 + 0] = 0.0; 
    rotation[1 * 4 + 1] = 1.0; 
    rotation[1 * 4 + 2] = 0.0; 
    rotation[1 * 4 + 3] = 0.0; 
    rotation[2 * 4 + 0] = 0.0; 
    rotation[2 * 4 + 1] = 0.0; 
    rotation[2 * 4 + 2] = 1.0; 
    rotation[2 * 4 + 3] = 0.0; 
    rotation[3 * 4 + 0] = 0.0; 
    rotation[3 * 4 + 1] = 0.0; 
    rotation[3 * 4 + 2] = 0.0; 
    rotation[3 * 4 + 3] = 1.0; 

	//initialize the parameters
	
	isLeftMouseActive      = false;
    isRightMouseActive     = false; 
    isMiddleMouseActive    = false; 
    oldMouseX   = 0; 
    oldMouseY   = 0; 
    mouseMove = false;
	scale     = 1.0;
    trans[0]    = 0.0;
    trans[1]    = 0.0;

    // Reset the rotation
    Reset(); 
}

TrackBall::~TrackBall()
{
	delete[] rotation;
}

void TrackBall::Reset()
{
	scale = 1.0;
    trans[0] = trans[1] = 0.0;
	offset[0] = offset[1] = offset[2] = 0.5;

    qx = qy = qz = 0.0; 
    qw = 1.0; 

    ComputeRotationMatrix(); 
}

void TrackBall::ResetOffset()
{
	offset[0] = offset[1] = offset[2] = 0.5;
}

void TrackBall::SetScale(float s)
{
	scale = s;
}
void TrackBall::SetValidButton(bool left, bool right, bool middle)
{
	leftButton = left;
	rightButton = right;
	middleButton = middle;
}
void TrackBall::SetWindowSize(int w, int h) 
{
    width = w; 
    height = h; 
}

void TrackBall::SetRotationMatrix(int pos, float value)
{
	rotation[pos] = value;
}

float* TrackBall::GetRotationMatrix()
{
    return rotation; 
}

float* TrackBall::GetScalePtr()
{
	return &scale;
}
float* TrackBall::GetTransPtr()
{
	return trans;
}
double * TrackBall::GetOffsetPtr()
{
	return offset;
}
void TrackBall::ComputeRotationMatrix() 
{
    rotation[0 * 4 + 0] = 1 - 2 * qy * qy - 2 * qz * qz; 
    rotation[1 * 4 + 0] = 2 * qx * qy - 2 * qw * qz; 
    rotation[2 * 4 + 0] = 2 * qx * qz + 2 * qw * qy; 
    rotation[0 * 4 + 1] = 2 * qx * qy + 2 * qw * qz; 
    rotation[1 * 4 + 1] = 1 - 2 * qx * qx - 2 * qz * qz; 
    rotation[2 * 4 + 1] = 2 * qy * qz - 2 * qw * qx; 
    rotation[0 * 4 + 2] = 2 * qx * qz - 2 * qw * qy; 
    rotation[1 * 4 + 2] = 2 * qy * qz + 2 * qw * qx; 
    rotation[2 * 4 + 2] = 1 - 2 * qx * qx - 2 * qy * qy; 
}

void TrackBall::ptov(int x, int y, double v[3])
{
    double d, a; 

    v[0] = (2.0 * x - width) / width; 
    v[1] = (height - 2.0 * y) / height; 
    d = sqrt(v[0] * v[0] + v[1] * v[1]); 
    v[2] = cos((PI/2.0) * ((d < 1.0) ? d : 1.0)); 
    a = 1.0 / sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]); 
    v[0] *= a; v[1] *= a; v[2] *= a; 
}

void TrackBall::StartDrag(int x, int y)
{
    oldX = x; oldY = y; 
    ox = qx; oy = qy; oz = qz; ow = qw; 
}

void TrackBall::Drag(int x, int y)
{
    double oldPos[3], newPos[3], dPos[3]; 
    double rx, ry, rz, rw; 
    double angle, sin_angle, axis[3], len; 

    // Project points onto the sphere
    ptov(oldX, oldY, oldPos); 
    ptov(x, y, newPos); 

    dPos[0] = newPos[0] - oldPos[0]; 
    dPos[1] = newPos[1] - oldPos[1]; 
    dPos[2] = newPos[2] - oldPos[2]; 

    if (dPos[0] != 0 || dPos[1] != 0 || dPos[2] != 0) {
        // Rotation angle and axis
        axis[0] = oldPos[1] * newPos[2] - oldPos[2] * newPos[1]; 
        axis[1] = oldPos[2] * newPos[0] - oldPos[0] * newPos[2]; 
        axis[2] = oldPos[0] * newPos[1] - oldPos[1] * newPos[0]; 
        angle = sqrt(dPos[0] * dPos[0] + dPos[1] * dPos[1] + dPos[2] * dPos[2]);                 

        // Rotation Quaternion
        sin_angle = sin(angle); 
        rx = axis[0] * sin_angle; 
        ry = axis[1] * sin_angle; 
        rz = axis[2] * sin_angle; 
        rw = cos(angle); 

        // Concatenate with the old rotation
        qx = ry * oz - rz * oy + rx * ow + ox * rw; 
        qy = rz * ox - rx * oz + ry * ow + oy * rw; 
        qz = rx * oy - ry * ox + rz * ow + oz * rw; 
        qw = rw * ow - rx * ox - ry * oy - rz * oz; 

        len = sqrt(qx * qx + qy * qy + qz * qz + qw * qw); 

        qx /= len; qy /= len; qz /= len; qw /= len; 

        ComputeRotationMatrix(); 
    }
}

void TrackBall::Rotate(double axis_x, double axis_y, double axis_z, double angle)
{
    double rx, ry, rz, rw; 
    double sin_angle, len; 

    // Rotation Quaternion
    sin_angle = sin(angle); 
    rx = axis_x * sin_angle; 
    ry = axis_y * sin_angle; 
    rz = axis_z * sin_angle; 
    rw = cos(angle); 

    // Concatenate with the old rotation
    ox = qx; oy = qy; oz = qz; ow = qw; 
    qx = ry * oz - rz * oy + rx * ow + ox * rw; 
    qy = rz * ox - rx * oz + ry * ow + oy * rw; 
    qz = rx * oy - ry * ox + rz * ow + oz * rw; 
    qw = rw * ow - rx * ox - ry * oy - rz * oz; 

    len = sqrt(qx * qx + qy * qy + qz * qz + qw * qw); 

    qx /= len; qy /= len; qz /= len; qw /= len; 

    ComputeRotationMatrix(); 
}
bool TrackBall::IsMouseDragging()
{
	return mouseMove;
}
void TrackBall::ReceiveMouseFunc(int button, int state, int x, int y)
{	
	if(!leftButton && button == GLUT_LEFT_BUTTON) return;
	if(!rightButton && button == GLUT_RIGHT_BUTTON) return;
	if(!middleButton && button == GLUT_MIDDLE_BUTTON) return;
	int modifiers = glutGetModifiers();
    if (state == GLUT_UP)
    {
	    switch (button)
        {
	        case GLUT_LEFT_BUTTON:
			    isLeftMouseActive = false;
                isMiddleMouseActive = false; 
    		    break;
            case GLUT_RIGHT_BUTTON:
                isRightMouseActive = false; 
                break; 
            case GLUT_MIDDLE_BUTTON:
                isMiddleMouseActive = false; 
                break; 
        }
	    mouseMove = false;
    }
    if (state == GLUT_DOWN)
    {
	    oldMouseX = x;
	    oldMouseY = y;

	    switch (button)
	    {
	    case GLUT_LEFT_BUTTON:
            if ((modifiers & GLUT_ACTIVE_SHIFT) > 0)
			    isMiddleMouseActive = true;
            else {
		        isLeftMouseActive = true;
                StartDrag(x, y); 
            }
		    break;
	    case GLUT_RIGHT_BUTTON:
		    isRightMouseActive = true; 
		    break;
	    case GLUT_MIDDLE_BUTTON:
		    isMiddleMouseActive = true;
		    break;
	    }
		
		//store the focus point
		double modelview[16],projection[16];
		int viewport[4];
		float winX,winY,winZ;
		double object_x = 0,object_y = 0,object_z = 0;    
		glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
		glGetDoublev(GL_PROJECTION_MATRIX, projection);
		glGetIntegerv(GL_VIEWPORT, viewport);
		winX=(float)x;
		winY=(float)viewport[3]-(float)y;
		glReadPixels(x, int(winY),1,1,GL_DEPTH_COMPONENT,GL_FLOAT, &winZ);
		gluUnProject((GLdouble)winX,(GLdouble)winY,(GLdouble)winZ,modelview,projection,viewport,
			&click[0],&click[1],&click[2]);
		click[0] = std::max(click[0], 0.0);
		click[1] = std::max(click[1], 0.0);
		click[2] = std::max(click[2], 0.0);
		click[0] = std::min(click[0], 1.0);
		click[1] = std::min(click[1], 1.0);
		click[2] = std::min(click[2], 1.0);
		//printf("(%.3f, %.3f, %.3f)\n", click[0], click[1], click[2]);

		//perfect way: use this click->offset and update rotation matrix
    }	
}
void TrackBall::ReceiveMouseMotionFunc(int x, int y)
{
	double temp = 512.0;
	if (isLeftMouseActive) {
	    mouseMove = true;
        Drag(x, y); 
    } 
    else if (isMiddleMouseActive) {
	    mouseMove = true;
		trans[0] += double(x - oldMouseX)*0.003; 
		trans[1] -= double(y - oldMouseY)*0.003;  
    }
    else if (isRightMouseActive) {
	    mouseMove = true;
        scale += (y - oldMouseY) * scale / 400.0;
    } 

    oldMouseX = x; oldMouseY = y;
	/*
	if(isLeftMouseActive)
	{
		//update this->trans using this->click
		this->offset[0] = this->offset[0] * this->offsetRatio + this->click[0] * (1.0-this->offsetRatio);
		this->offset[1] = this->offset[1] * this->offsetRatio + this->click[1] * (1.0-this->offsetRatio);
		this->offset[2] = this->offset[2] * this->offsetRatio + this->click[2] * (1.0-this->offsetRatio);
	}
	if(isRightMouseActive)
	{
		this->offset[0] = this->offset[0] * this->offsetRatio + this->click[0] * (1.0-this->offsetRatio);
		this->offset[1] = this->offset[1] * this->offsetRatio + this->click[1] * (1.0-this->offsetRatio);
		this->offset[2] = this->offset[2] * this->offsetRatio + this->click[2] * (1.0-this->offsetRatio);
	}*/
}