/*
Author: Cao Thanh Tung & Gao Mingcen
Date: 28/02/2013

File Name: TrackBall.h

Class Definition of TrackBall, a controller of 3D space for visualization

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#ifndef TRACKBALL_H
#define TRACKBALL_H

class TrackBall
{	
private:
    int width, height; 
    int oldX, oldY; 
	double qx, qy, qz, qw; 
    double ox, oy, oz, ow; 
    float *rotation; 
	
	//outside operation	
	bool isLeftMouseActive;
    bool isRightMouseActive; 
    bool isMiddleMouseActive; 
    int oldMouseX; 
    int oldMouseY; 
    bool mouseMove;
	float scale;
	float trans[2];
	double click[3];
	double offset[3];
	double offsetRatio;

	//button
	bool leftButton;
	bool middleButton;
	bool rightButton;

public:
	TrackBall();
	TrackBall(int w, int h);
	TrackBall(int w, int h, bool left, bool right, bool middle);
	~TrackBall();

	//self-maintained function
	void	Reset();
	void	ResetOffset();
	void	SetScale(float s);
	void	SetValidButton(bool left, bool right, bool middle);
	void    SetWindowSize(int w, int h);
    void    StartDrag(int x, int y); 
    void    Drag(int x, int y); 
	void	SetRotationMatrix(int pos, float value);
    float*  GetRotationMatrix(); 
	float*	GetScalePtr();
	float*	GetTransPtr();
	double * GetOffsetPtr();
    void    Rotate(double axis_x, double axis_y, double axis_z, double angle);
	bool	IsMouseDragging();

	//receive signals from uses function
	void	ReceiveMouseFunc(int button, int state, int x, int y);
	void	ReceiveMouseMotionFunc( int x, int y);
	/*
	friend void glutDisplay();
	friend void glutReshape(int width,int height);
	friend void glutMouse(int button, int state, int x, int y);
	friend void glutMouseMotion(int x, int y);
	friend void glutKeyboard(unsigned char key, int x, int y);
	friend void callback(int control);
	friend class ShapeRenderer;*/

private:
    void    ptov(int x, int y, double p[3]); 
    void    ComputeRotationMatrix(); 
};

#endif